@@ -1,5 +1,4 @@
 /*
-
 *************************************************************************
 
 ArmageTron -- Just another Tron Lightcycle Game in 3D.
@@ -43,17 +42,31 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "tMath.h"
 #include "nConfig.h" 
 
+#include "tRecorder.h"
+
 #include <vector>
 #include <set>
 
 #define SMALL_FLOAT 1E-30
 const REAL se_maxGridSize = 1E+15;	// this gives us way more space than OpenGL allows us to render
 
+// #define EPS 1E-2
+
 int se_debugExt=0;
 
-tDEFINE_REFOBJ( ePoint );
-tDEFINE_REFOBJ( eHalfEdge );
-tDEFINE_REFOBJ( eFace );
+// counts how many edges we should attempt to simplify
+static int se_simplifyEdges = 0;
+
+// after one edge has been successfully simplified, simplify that many more,
+// it seems to pay
+static const int se_simplifyEdgesSuccess = 20;
+
+// after one edge has been inserted, try to simplify that many old ones
+static const int se_simplifyEdgesNew = 2;
+
+tDEFINE_REFOBJ( ePoint )
+tDEFINE_REFOBJ( eHalfEdge )
+tDEFINE_REFOBJ( eFace )
 
 
 REAL se_EstimatedRangeOfMult( const eCoord &a,const eCoord &b )
@@ -62,6 +75,10 @@ REAL se_EstimatedRangeOfMult( const eCoord &a,const eCoord &b )
     return fabs( b*a ) + fabs( b*( a.Turn( turn ) ) );
 }
 
+REAL st_GetDifference( const eCoord &a,const eCoord &b )
+{
+    return (a-b).NormSquared();
+}
 
 eCoord    eTempEdge::Vec()         const{return halfEdges[0]->Vec();}
 eCoord&   eTempEdge::Coord(int i)  const{return *halfEdges[i]->Point();}
@@ -187,10 +204,19 @@ inline void eHalfEdge::Unlink()
 
 void eGrid::Check() const{
     //    return;
-#ifdef DEBUG
+    if ( a )
+    {
+        tASSERT( a->Point() == B );
+        tASSERT( a->other->Point() == C );
+        tASSERT( b->Point() == C );
+        tASSERT( b->other->Point() == A );
+        tASSERT( c->Point() == A );
+        tASSERT( c->other->Point() == B );
+    }
+
+#ifdef DEBUG_EXPENSIVE
     if (!doCheck)
         return;
-#endif
 
     int i;
     for (i = points.Len()-1; i>=0; i--)
@@ -201,6 +227,7 @@ void eGrid::Check() const{
 
     for (i = edges.Len()-1; i>=0; i--)
         edges(i)->Check();
+#endif
 }
 
 
@@ -322,14 +349,20 @@ void se_LinkFaces( tControlledPTR< eFace >& old, eFace* replacement )
 typedef std::pair< eFace*, REAL > eFaceScorePair;
 typedef std::set< const eFace* > eFaceVisitedSet;
 
-eFaceScorePair se_FindBestReplacement( const eFace *old, const eCoord& coord, eFaceVisitedSet& visited )
+struct eFaceReplacementArgument
+{
+    eCoord coord, direction, lastDirection;
+    eFaceVisitedSet visited;
+};
+
+eFaceScorePair se_FindBestReplacement( const eFace *old, eFaceReplacementArgument& arg )
 {
     // return invalid return if the face was already visited
-    if ( visited.find( old ) != visited.end() )
+    if ( arg.visited.find( old ) != arg.visited.end() )
         return eFaceScorePair( 0, -se_maxGridSize );
 
     // register face as visited
-    visited.insert( old );
+    arg.visited.insert( old );
 
     // find entry in replacement map
 
@@ -355,12 +388,12 @@ eFaceScorePair se_FindBestReplacement( const eFace *old, const eCoord& coord, eF
             {
                 // enter it directly
                 current.first = face;
-                current.second = face->Insideness( coord );
+                current.second = face->Insideness( arg.coord, arg.direction, arg.lastDirection );
             }
             else
             {
                 // no, we have to look for a replacement recursively
-                current = se_FindBestReplacement( face, coord, visited );
+                current = se_FindBestReplacement( face, arg );
             }
 
             if ( current.second > best.second && current.first )
@@ -379,16 +412,46 @@ eReplacementStorage& eFace::GetReplacementStorage() const
     return *replacementStorage;
 }
 
+eFace* eFace::FindReplacement( const eCoord& coord, const eCoord& direction, const eCoord& lastDirection ) const
+{
+    // check if we don't need an replacement
+    tASSERT( !this->IsInGrid() );
+
+    // set of already visited faces
+    eFaceReplacementArgument arg;
+    arg.coord = coord;
+    arg.direction = direction;
+    arg.lastDirection = lastDirection;
+
+    // delegate search to recursive function
+    return se_FindBestReplacement( this, arg ).first;
+}
+
+eFace* eFace::FindReplacement( const eCoord& coord, const eCoord& direction ) const
+{
+    // check if we don't need an replacement
+    tASSERT( !this->IsInGrid() );
+
+    // set of already visited faces
+    eFaceReplacementArgument arg;
+    arg.coord = coord;
+    arg.direction = direction;
+
+    // delegate search to recursive function
+    return se_FindBestReplacement( this, arg ).first;
+}
+
 eFace* eFace::FindReplacement( const eCoord& coord ) const
 {
     // check if we don't need an replacement
     tASSERT( !this->IsInGrid() );
 
     // set of already visited faces
-    eFaceVisitedSet visited;
+    eFaceReplacementArgument arg;
+    arg.coord = coord;
 
     // delegate search to recursive function
-    return se_FindBestReplacement( this, coord, visited ).first;
+    return se_FindBestReplacement( this, arg ).first;
 }
 
 eFace::~eFace()
@@ -490,7 +553,7 @@ void eFace::Unlink(){
 }
 
 // ckeck if the given point lies inside this face; return value positive if it is inside and negative if outside
-REAL eFace::Insideness(const eCoord &c) const
+REAL eFace::Insideness(const eCoord& pos, const eCoord& direction, const eCoord& lastDirection ) const
 {
     // TODO: bounding box check
 
@@ -511,11 +574,18 @@ REAL eFace::Insideness(const eCoord &c) const
         tASSERT(run && run->Face() == this);
         eCoord *p = run->Point();
         eCoord ed=*last_p - *p;
-        eCoord cc= c - *p;
+        eCoord cc= pos - *p;
 
         REAL val = ed*cc;
+        // REAL tol = 10 * EPS/(ed.Norm() + EPS*1000);
+        REAL tol = 10 * EPS;
+        REAL retDir = direction*ed;
+        REAL retLastDir = lastDirection*ed;
+        val += tol*(retDir + tol*retLastDir);
         if ( val < ret )
+        {
             ret = val;
+        }
 
         last_p = p;
         run = run->next;
@@ -525,6 +595,16 @@ REAL eFace::Insideness(const eCoord &c) const
     return ret;
 }
 
+REAL eFace::Insideness(const eCoord& c, const eCoord& direction ) const
+{
+    return Insideness( c, direction, eCoord() );
+}
+
+REAL eFace::Insideness(const eCoord& c ) const
+{
+    return Insideness( c, eCoord(), eCoord() );
+}
+
 // ckeck if the given point lies inside this face (edges included)
 bool eFace::IsInside(const eCoord &c) const
 {
@@ -571,31 +651,76 @@ bool eFace::IsInside(const eCoord &c) const
 void eCoord::Print(std::ostream &s)const{ s << "(" << x << "," << y << ")"; }
 void eHalfEdge::Print(std::ostream &s)const{ s << "[" << *Point() << "->" << *other->Point() << "]"; }
 void eFace::Print(std::ostream &s)const{ s << "[" << *edge->Point() << "->" << *edge->next->Point() << "->" << *edge->next->next->Point() << "]"; }
+void eCoord::Read(std::istream &s)
+{
+    char c;
+    s >> c;
+    tASSERT( c == '(' );
+    s >> x;
+    s >> c;
+    tASSERT( c == ',' );
+    s >> y;
+    s >> c;
+    tASSERT( c == ')' );
+}
 
 
 // **************************
 //          eGrid
 // **************************
 
-
-int eGrid::DirectionWinding(const eCoord&dir) const
+/*
+ * Return the winding number closest to direction dir
+ */
+int eGrid::DirectionWinding(const eCoord&dir)
 {
-    if (dir.x >= fabs(dir.y))
-        return 0;
-    if (dir.y >= fabs(dir.x))
-        return 3;
-    if (-dir.x >= fabs(dir.y))
-        return 2;
-    if (-dir.y >= fabs(dir.x))
-        return 1;
-    return -1;
+    return axis.NearestWinding(dir);
 }
 
+/*
+ * Return the direction associated with a winding number
+ */
+eCoord eGrid::GetDirection(int winding)
+{
+    return axis.GetDirection(winding);
+}
+
+/*
+ * Define the number of winding to be used on this grid
+ */
+void eGrid::SetWinding(int number)
+{
+    axis.Init(number);
+}
+
+/*
+ * Define the number of winding and their values.
+ * When the normalize flag is set, the values are normalized.
+ */
+void eGrid::SetWinding(int number, eCoord directions[], bool normalise)
+{
+    axis.Init(number, directions, normalise);
+}
+
+/*
+ * change/turn currentWinding by 'direction' steps
+ */
+void eGrid::Turn(int &currentWinding, int direction)
+{
+    axis.Turn(currentWinding, direction);
+}
+
+
 static eHalfEdge *leakEdge =NULL;
 static ePoint    *leakPoint=NULL;
 
+static int se_drawLineTimeout = 10000;
+static tSettingItem<int> se_drawLineTimeoutConf("DRAWLINE_TIMEOUT", se_drawLineTimeout);
 
 ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change_grid){
+    // for every edge we add, allow to simplify
+    se_simplifyEdges += se_simplifyEdgesNew;
+
     // prepare a teporary edge so we always know what the wall we are placing looks like
     tStackObject< eTempEdge > toBePlaced( *start, end, w );
     //tJUST_CONTROLLED_PTR< eWall > wal( w );
@@ -626,11 +751,17 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
 
 #ifdef DEBUG
     bool foundCurrentEdge;
+    static const int laststarts_max = 10;
+    ePoint *laststarts[laststarts_max];
 #endif
 
     eCoord dir=eCoord(1,1); // the direction we are going
 
-    int timeout = 10000;
+    int timeout = se_drawLineTimeout;
+
+    // distance to end to know when we are going backwards
+    REAL distToEnd = ( *start - end ).NormSquared();
+    ePoint * lastStart = start;
 
     bool goon=1;
     while (goon && dir.NormSquared()>0){
@@ -641,13 +772,21 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
         //      st_Breakpoint();
 #endif
 
-
         dir=end-*start; // the direction we are going
 
-#ifdef DEBUG
-        ePoint *laststarts[10];
+        // check if we got closer to the end, if not, accelerate timeout and set breakpoint
+        REAL newDistToEnd = dir.NormSquared();
+        // std::cout << distToEnd << ", " << newDistToEnd << "\n";
+        if ( newDistToEnd > distToEnd && start != lastStart )
+        {
+            //st_Breakpoint();
+            timeout -= 100;
+        }
+        distToEnd = newDistToEnd;
+        lastStart = start;
 
-        for (int i = 9; i>=1; i--)
+#ifdef DEBUG
+        for (int i = laststarts_max-1; i>=1; i--)
             laststarts[i] = laststarts[i-1];
 
         laststarts[0] = start;
@@ -724,7 +863,7 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                 if (right_test < 0 && left_test < 0)
                     score *= -1;     // here, - * - still is -.....
 
-                if (!best || score > best_score  && rvec * lvec < 0
+                if (!best || ( score > best_score  && rvec * lvec < 0 )
                    )
                     best = run;
 
@@ -829,10 +968,17 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                 ePoint*    C=b->Point();
 
                 if ((*B) == end)
+                {
+                    c->SetWall(toBePlaced.Wall());
                     return B;
+                }
 
                 if ((*C) == end)
+                {
+                    tASSERT( b->other );
+                    b->other->SetWall(toBePlaced.Wall());
                     return C;
+                }
 
                 if(!exactly_onLeft)
                 {
@@ -923,9 +1069,9 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
 
                     tASSERT(b->other);
 
-                    eFace *G=b->other->face;
+                    // eFace *G=b->other->face;
 
-                    tASSERT(G);
+                    // tASSERT(G);
 
                     eHalfEdge *e1=b->other->next;
                     eHalfEdge *e2=e1->next;
@@ -1144,10 +1290,10 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                                  bb.NormSquared() + cc.NormSquared();
 
                     tASSERT(scale > 0);
-                    tASSERT(aa.NormSquared() > scale *SMALL_FLOAT);
-                    tASSERT(bb.NormSquared() > scale *SMALL_FLOAT);
-                    tASSERT(dd.NormSquared() > scale *SMALL_FLOAT);
-                    tASSERT(cc.NormSquared() > scale *SMALL_FLOAT);
+                    //tASSERT(aa.NormSquared() > scale *SMALL_FLOAT);
+                    //tASSERT(bb.NormSquared() > scale *SMALL_FLOAT);
+                    //tASSERT(dd.NormSquared() > scale *SMALL_FLOAT);
+                    //tASSERT(cc.NormSquared() > scale *SMALL_FLOAT);
 
 #ifdef DEBUG
                     Check();
@@ -1308,7 +1454,7 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                             // place the first bit of wall
                             cc->SetWall(wa);
 
-                            // modify the temp edge to be placed to repesent the rest of the wall
+                            // modify the temp edge to be placed to represent the rest of the wall
                             toBePlaced.SetWall( wb );
                             toBePlaced.Coord(0) = *E;
                         }
@@ -1371,8 +1517,9 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                             dd = a->other;
                             bb = b->other;
                             aeq = true;
-                        }
 
+                            se_LinkFaces( oldG, oldF );
+                        }
 
                         if (*C == *E)
                         {
@@ -1392,6 +1539,8 @@ ePoint * eGrid::DrawLine(ePoint *start, const eCoord &end, eWall *w, bool change
                             dd = d;
                             bb = c;
                             ceq = true;
+
+                            se_LinkFaces( oldF, oldG );
                         }
 
                         if (!ceq)
@@ -1683,7 +1832,6 @@ bool eFace::CorrectArea()
     return newarea > area;
 }
 
-
 bool eHalfEdge::Simplify(eGrid *grid)
 {
     if (GetWall() && GetWall()->Deletable() && face)
@@ -1691,6 +1839,11 @@ bool eHalfEdge::Simplify(eGrid *grid)
         ClearWall();
     }
 
+    if (other && other->GetWall() && other->GetWall()->Deletable() && other->face)
+    {
+        other->ClearWall();
+    }
+
     if (GetWall() || !other || other->GetWall() || !face || !other->face)
         return false;
 
@@ -1799,11 +1952,14 @@ bool eHalfEdge::Simplify(eGrid *grid)
 
     tControlledPTR< eHalfEdge > keep( this );
 
+    int idrec = ID;
+    tRecorderSync< int >::Archive( "_GRID_SIMPLIFY_REAL", 8, idrec );
+
     // the faces that will be removed
     tControlledPTR< eFace > killed1 = grid->ZombifyFace( face );
     tControlledPTR< eFace > killed2 = grid->ZombifyFace( other->face );
 
-    // all faces that will replace the to get added to their replacement list
+    // all faces that will replace them get added to their replacement list
     run = this->other->next->other->next;
     do{
         tASSERT(run->other);
@@ -1857,7 +2013,20 @@ void eGrid::SimplifyNum(int e){
 #ifdef DEBUG
         eHalfEdge *Esave=E;
 #endif
-        E->Simplify(this);
+        if ( !E || !E->other )
+            return;
+
+        // this point will be removed. Make sure it is none of the points on the outer rim.
+        ePoint * toBeRemoved = E->other->Point();
+        if ( toBeRemoved == A || toBeRemoved == B || toBeRemoved == C )
+            return;
+
+        tRecorderSync< int >::Archive( "_GRID_SIMPLIFY", 8, e );
+
+        if ( E->Simplify(this) )
+        {
+            se_simplifyEdges += se_simplifyEdgesSuccess;
+        }
 
 #ifdef DEBUG
         tASSERT(Esave);
@@ -1880,7 +2049,10 @@ void eGrid::SimplifyAll(int count){
 
     static int counter=0;
 
-    for(;count>0;count--){
+    // try to simplify at least one edge
+    se_simplifyEdges++;
+
+    for(;count>0 && se_simplifyEdges>0;count--,se_simplifyEdges--){
         counter--;
 
         if (counter<0 || counter>=edges.Len())
@@ -1911,8 +2083,8 @@ void eGrid::Create(){
 
 
     A=tNEW(ePoint) (base.Turn(1,0));
-    B=tNEW(ePoint) (base.Turn(-.5,.7));
-    C=tNEW(ePoint) (base.Turn(-.5,-.7));
+    B=tNEW(ePoint) (base.Turn(-.5,.87));
+    C=tNEW(ePoint) (base.Turn(-.5,-.87));
 
     a=tNEW(eHalfEdge) (B,C,tNEW(eWall(this)));
     b=tNEW(eHalfEdge) (C,A,tNEW(eWall(this)));
@@ -1939,6 +2111,8 @@ void eGrid::Create(){
 void eGrid::Clear(){
     eHalfEdge::ClearPathData();
 
+    eGameObject::DeleteAll( this );
+
     base=eCoord(0,100);
     maxNormSquared=0;
 
@@ -1983,8 +2157,8 @@ void eGrid::Grow()
     c->ClearWall();
 
     ePoint *newA=tNEW(ePoint) (base.Turn(1,0));
-    ePoint *newB=tNEW(ePoint) (base.Turn(-.5,.7));
-    ePoint *newC=tNEW(ePoint) (base.Turn(-.5,-.7));
+    ePoint *newB=tNEW(ePoint) (base.Turn(-.5,.87));
+    ePoint *newC=tNEW(ePoint) (base.Turn(-.5,-.87));
 
     eHalfEdge *newa=tNEW(eHalfEdge) (newB,newC,tNEW(eWall(this)));
     eHalfEdge *newb=tNEW(eHalfEdge) (newC,newA,tNEW(eWall(this)));
@@ -2030,7 +2204,9 @@ void eGrid::Range(REAL NormSquared){
         Create();
     //    tERR_ERROR_INT("Wanted to insert something into the nonexistant grid!");
 
-    while (NormSquared > maxNormSquared *.1 && maxNormSquared < se_maxGridSize )
+    // the magic factor of 4 is chosen so that no points can be added outside of the
+    // current tiangle abc added by Grow().
+    while (NormSquared * 4 > maxNormSquared && maxNormSquared < se_maxGridSize )
         Grow();
 }
 
@@ -2100,10 +2276,21 @@ eCoord eHalfEdge::IntersectWithCareless(const eHalfEdge *e2) const
     eCoord e2_t = *Point()     - *e2->Point();
 
     REAL inv = (vec*e2vec);
-    if (inv!=0)
+    if (fabs(inv) > EPS * se_EstimatedRangeOfMult( vec, e2vec ) )
+    {
+        // return intersection poin
         return (*Point()-vec*((e2_t*e2vec)/inv));
+    }
     else
-        return (*Point()-vec*((e2_t*e2vec)));
+    {
+        // lines lie almost parallel; return "inverse center of gravity" ( the swapped
+        // weighting is intentional, the smaller line should have a bigger influence )
+        REAL m1 = vec.Norm();
+        REAL m2 = e2vec.Norm();
+        if ( m1 + m2 <= 0 ) // sanity check weights
+            m1=m2=1;
+        return ((*Point() + *other->Point())*m2 + (*e2->Point() + *e2->other->Point())*m1)*(.5/(m1+m2));
+    }
 }
 
 // the same, but with checks: if the two edges don't intersect,
@@ -2207,6 +2394,9 @@ void eGrid::AddEdge    (eHalfEdge *e)
     tASSERT(e->Other() && *e->Point() != *e->Other()->Point());
 
     edges.Add(e, e->ID);
+
+    int idrec = e->ID;
+    tRecorderSync< int >::Archive( "_GRID_ADD_EDGE", 8, idrec );
 }
 
 void eGrid::RemoveEdge (eHalfEdge *e)
@@ -2217,6 +2407,9 @@ void eGrid::RemoveEdge (eHalfEdge *e)
     if (e->ID < 0)
         return;
 
+    int idrec = e->ID;
+    tRecorderSync< int >::Archive( "_GRID_REMOVE_EDGE", 8, idrec );
+
     edges.Remove(e, e->ID);
 }
 
@@ -2375,19 +2568,22 @@ eFace * eGrid::FindSurroundingFace(const eCoord &pos, eFace *currentFace) const{
                 continue;
             }
 
-            eCoord vec=pos - (*run->Point()); // the vector to our destination
+            if ( run->Other() && run->Other()->Face() )
+            {
+                eCoord vec=pos - (*run->Point()); // the vector to our destination
 
-            REAL score = run->Vec() * vec;
+                REAL score = run->Vec() * vec;
+
+                if (score > bestScore)
+                {
+                    best      = run;
+                    bestScore = score;
+                }
+            }
 
             if (!first)
                 first = run;
 
-            if (score > bestScore)
-            {
-                best      = run;
-                bestScore = score;
-            }
-
             run = run->Next();
         }
 
