@@ -25,10 +25,11 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 */
 
+#include "eSound.h"
 #include "config.h"
 #include "tMemManager.h"
 #include "tDirectories.h"
-#include "eSound.h"
+#include "tRandom.h"
 #include "tError.h"
 #include <string>
 #include "tConfiguration.h"
@@ -38,6 +39,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include <iostream>
 #include <stdlib.h>
 #include "eGrid.h"
+#include "tException.h"
 
 //eGrid* eSoundPlayer::S_Grid = NULL;
 
@@ -89,7 +91,11 @@ void fill_audio(void *udata, Uint8 *stream, int len)
     int i;
     if (eGrid::CurrentGrid())
         for(i=eGrid::CurrentGrid()->Cameras().Len()-1;i>=0;i--)
-            eGrid::CurrentGrid()->Cameras()(i)->SoundMix(stream,len);
+        {
+            eCamera *pCam = eGrid::CurrentGrid()->Cameras()(i);
+            if(pCam)
+                pCam->SoundMix(stream,len);
+        }
 
     for(i=se_globalPlayers.Len()-1;i>=0;i--)
         se_globalPlayers(i)->Mix(stream,len,0,1,1);
@@ -107,15 +113,60 @@ void fill_audio(void *udata, Uint8 *stream, int len)
 #endif
 }
 
+#ifndef DEDICATED
+#ifdef DEFAULT_SDL_AUDIODRIVER
+
+// stringification, yep, two levels required
+#define XSTRING(s) #s
+#define STRING(s) XSTRING(s)
+
+// call once to initialize SDL sound subsystem
+static bool se_SoundInitPrepare()
+{
+    // initialize audio subsystem with predefined, hopefully good, driver
+    if ( ! getenv("SDL_AUDIODRIVER") ) {
+        char * arg = "SDL_AUDIODRIVER=" STRING(DEFAULT_SDL_AUDIODRIVER);
+        putenv(arg);
+
+        if ( SDL_InitSubSystem(SDL_INIT_AUDIO) >= 0 )
+            return true;
+
+        putenv("SDL_AUDIODRIVER=");
+    }
+
+    // if that fails, try what the user wanted
+    return ( SDL_InitSubSystem(SDL_INIT_AUDIO) >= 0 );
+}
+#endif
+#endif
 
 void se_SoundInit()
 {
 #ifndef DEDICATED
+    // save configuration file with sound disabled on first use so we don't try again
+    bool needSave = false;
+    static bool firstRun = true;
+    if ( st_FirstUse )
     {
-        static bool first = true;
-        if ( first && !SDL_WasInit( SDL_INIT_AUDIO ) )
+        needSave = true;
+        int sound_quality_back = sound_quality;
+        sound_quality = SOUND_OFF;
+        st_SaveConfig();
+        if ( firstRun )
+            con << tOutput("$sound_firstinit");
+        sound_quality=sound_quality_back;
+    }
+
+    if ( sound_quality != SOUND_OFF )
+    {
+#ifdef DEFAULT_SDL_AUDIODRIVER
+        static bool init = se_SoundInitPrepare();
+        if ( !init )
             return;
-        first = false;
+#endif
+        if ( firstRun && !SDL_WasInit( SDL_INIT_AUDIO ) )
+            return;
+        firstRun = false;
     }
 
     if (!sound_is_there && sound_quality!=SOUND_OFF)
@@ -142,15 +193,6 @@ void se_SoundInit()
         desired.callback = fill_audio;
         desired.userdata = NULL;
 
-        // save configuration file with sound disabled
-        int sound_quality_back = sound_quality;
-        if ( st_FirstUse )
-        {
-            sound_quality = SOUND_OFF;
-            st_SaveConfig();
-            con << tOutput("$sound_firstinit");
-        }
-
 #ifdef HAVE_LIBSDL_MIXER
         uses_sdl_mixer=true;
 
@@ -182,14 +224,6 @@ void se_SoundInit()
         uses_sdl_mixer=false;
         sound_is_there=(SDL_OpenAudio(&desired,&audio)>=0);
 #endif
-
-        // restore sound settings
-        if ( sound_quality != sound_quality_back )
-        {
-            sound_quality = sound_quality_back;
-            st_SaveConfig();
-        }
-
         if (sound_is_there && (audio.format!=AUDIO_S16SYS || audio.channels!=2))
         {
             uses_sdl_mixer=false;
@@ -206,25 +240,32 @@ void se_SoundInit()
         {
             //for(int i=wavs.Len()-1;i>=0;i--)
             //wavs(i)->Init();
+#ifdef DEBUG
             tOutput o;
             o.SetTemplateParameter(1,audio.freq);
             o.SetTemplateParameter(2,audio.samples);
             o << "$sound_inited";
             con << o;
+#endif
+            se_SoundPause(false);
         }
     }
+
+    // save sound settings, they appear to work
+    if ( needSave )
+    {
+        st_SaveConfig();
+    }
 #endif
 }
 
 void se_SoundExit(){
 #ifndef DEDICATED
-    se_SoundLock();
+    eSoundLocker locker;
 
     eWavData::UnloadAll();
     se_SoundPause(true);
 
-    se_SoundUnlock();
-
     if (sound_is_there){
 #ifdef DEBUG
         con << tOutput("$sound_disabling");
@@ -292,25 +333,29 @@ void se_SoundPause(bool p){
 eWavData* eWavData::s_anchor = NULL;
 
 eWavData::eWavData(const char * fileName,const char *alternative)
-        :tListItem<eWavData>(s_anchor),data(NULL),len(0),freeData(false){
+        :tListItem<eWavData>(s_anchor),data(NULL),len(0),freeData(false), loadError(false){
     //wavs.Add(this,id);
     filename     = fileName;
     filename_alt = alternative;
 
 }
 
-
 void eWavData::Load(){
     //wavs.Add(this,id);
 
     if (data)
+    {
+        loadError = false;
         return;
+    }
 
 #ifndef DEDICATED
 
+    static char const * errorName = "Sound Error";
 
     freeData = false;
 
+    loadError = true;
 
     alt=false;
 
@@ -325,7 +370,7 @@ void eWavData::Load(){
                 tOutput err;
                 err.SetTemplateParameter(1, filename);
                 err << "$sound_error_filenotfound";
-                tERR_ERROR(err);
+                throw tGenericException(err, errorName);
             }
             else
                 alt=true;
@@ -337,7 +382,7 @@ void eWavData::Load(){
                 tOutput err;
                 err.SetTemplateParameter(1, "sound/expl.waw");
                 err << "$sound_error_filenotfount";
-                tERR_ERROR(err);
+                throw tGenericException(err, errorName);
             }
             else
                 len=0;
@@ -362,7 +407,7 @@ void eWavData::Load(){
         SDL_AudioCVT cvt;
         if ( -1 == SDL_BuildAudioCVT( &cvt, spec.format, spec.channels, spec.freq, AUDIO_S16SYS, spec.channels, spec.freq ) )
         {
-            tERR_ERROR(err);
+            throw tGenericException(err, errorName);
         }
 
         cvt.buf=reinterpret_cast<Uint8 *>( malloc( len * cvt.len_mult ) );
@@ -373,7 +418,7 @@ void eWavData::Load(){
 
         if ( -1 == SDL_ConvertAudio( &cvt ) )
         {
-            tERR_ERROR(err);
+            throw tGenericException(err, errorName);
         }
 
         SDL_FreeWAV( data );
@@ -402,6 +447,8 @@ void eWavData::Load(){
     con << "at " << spec.freq << " Hz,\n";
 
     con << samples << " samples in " << len << " bytes.\n";
+
+    loadError = false;
 #endif
 #endif
 #endif
@@ -409,11 +456,12 @@ void eWavData::Load(){
 
 void eWavData::Unload(){
 #ifndef DEDICATED
+    loadError = false;
+
     //wavs.Add(this,id);
     if (data){
-        se_SoundLock();
+        eSoundLocker locker;
         if ( freeData )
-
         {
 
             free(data);
@@ -432,7 +480,6 @@ void eWavData::Unload(){
 
         data=NULL;
         len=0;
-        se_SoundUnlock();
     }
 #endif
 }
@@ -459,7 +506,14 @@ bool eWavData::Mix(Uint8 *dest,Uint32 playlen,eAudioPos &pos,
 #ifndef DEDICATED
     if ( !data )
     {
-        return false;
+        if( !loadError )
+        {
+            Load();
+        }
+        if ( !data )
+        {
+            return false;
+        }
     }
 
     playlen/=4;
@@ -685,8 +739,11 @@ void eWavData::Loop(){
 void eAudioPos::Reset(int randomize){
 #ifndef DEDICATED
     if (randomize){
-        fraction=int(SPEED_FRACTION*(rand()/float(RAND_MAX)));
-        pos=int(randomize*(rand()/float(RAND_MAX)));
+        tRandomizer & randomizer = tRandomizer::GetInstance();
+        fraction = randomizer.Get( SPEED_FRACTION );
+        // fraction=int(SPEED_FRACTION*(rand()/float(RAND_MAX)));
+        pos=randomizer.Get( randomize );
+        // pos=int(randomize*(rand()/float(RAND_MAX)));
     }
     else
         fraction=pos=0;
@@ -704,7 +761,11 @@ eSoundPlayer::eSoundPlayer(eWavData &w,bool l)
         goon[i]=true;
 }
 
-eSoundPlayer::~eSoundPlayer(){}
+eSoundPlayer::~eSoundPlayer()
+{
+    eSoundLocker locker;
+    se_globalPlayers.Remove(this,id);
+}
 
 bool eSoundPlayer::Mix(Uint8 *dest,
                        Uint32 len,
@@ -744,9 +805,8 @@ void eSoundPlayer::End(){
 void eSoundPlayer::MakeGlobal(){
     wav->Load();
 
-    se_SoundLock();
+    eSoundLocker locker;
     se_globalPlayers.Add(this,id);
-    se_SoundUnlock();
 }
 
 
@@ -828,3 +888,13 @@ void se_SoundMenu(){
     //  se_SoundPause(false);
 }
 
+eSoundLocker::eSoundLocker()
+{
+    se_SoundLock();
+}
+
+eSoundLocker::~eSoundLocker()
+{
+    se_SoundUnlock();
+}
+
