@@ -25,10 +25,10 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 */
 
+#include "uInput.h"
 #include "tMemManager.h"
 #include "rScreen.h"
 #include "tInitExit.h"
-#include "uInput.h"
 #include "tConfiguration.h"
 #include "rConsole.h"
 #include "uMenu.h"
@@ -45,27 +45,22 @@ uAction::uAction(uAction *&anchor,const char* name,
         :tListItem<uAction>(anchor),type(t),priority(priority_),internalName(name){
     globalID = localID = su_allActionsLen++;
 
-    int i;
-
     tASSERT(localID < uMAX_ACTIONS);
 
     su_allActions[localID] = this;
 
     tString descname;
     descname << "input_" << name << "_text";
-    for(i=descname.Len()-1;i>=0;i--)
-        descname[i]=tolower(descname[i]);
+    tToLower( descname );
 
-    ((tOutput&)description).AddLocale(descname);
+    const_cast<tOutput&>(description).AddLocale(descname);
 
     tString helpname;
     helpname << "input_" << name << "_help";
-    for(i=helpname.Len()-1;i>=0;i--)
-        helpname[i]=tolower(helpname[i]);
+    tToLower( helpname );
 
-    ((tOutput&)helpText).AddLocale(helpname);
-
-};
+    const_cast<tOutput&>(helpText).AddLocale(helpname);
+}
 
 uAction::uAction(uAction *&anchor,const char* name,
                  const tOutput& desc,
@@ -78,12 +73,21 @@ uAction::uAction(uAction *&anchor,const char* name,
     tASSERT(localID < uMAX_ACTIONS);
 
     su_allActions[localID] = this;
-};
+}
 
 uAction::~uAction(){
     su_allActions[localID] = NULL;
 }
 
+uAction * uAction::Find( char const * name )
+{
+    for(int i=su_allActionsLen-1;i>=0;i--)
+        if(!strcmp(name,su_allActions[i]->internalName))
+            return su_allActions[i];
+
+    return 0;
+}
+
 // ****************************************
 // a configuration class for keyboard binds
 // ****************************************
@@ -91,7 +95,7 @@ uAction::~uAction(){
 class tConfItem_key:public tConfItemBase{
 public:
     tConfItem_key():tConfItemBase("KEYBOARD"){}
-    ~tConfItem_key(){};
+    ~tConfItem_key(){}
 
     // write the complete keymap
     virtual void WriteVal(std::ostream &s){
@@ -122,13 +126,10 @@ public:
             s >> in;
             if (uBindPlayer::IsKeyWord(in))
             {
-                if ( keymap[keysym] )
+                keymap[keysym] = NULL;
+                keymap[keysym]=uBindPlayer::NewBind(s);
+                if (!keymap[keysym]->act)
                 {
-                    delete keymap[keysym];
-                }
-                keymap[keysym]=tNEW(uBindPlayer)(s);
-                if (!keymap[keysym]->act){
-                    delete keymap[keysym];
                     keymap[keysym]=NULL;
                 }
                 /* if (global_bind::IsKeyWord(in))
@@ -170,6 +171,7 @@ uActionPlayer *uActionPlayer::Find(int id){
     while (run){
         if (run->ID() == id)
             return static_cast<uActionPlayer*>(run);
+        run = run->Next();
     }
 
     return NULL;
@@ -212,8 +214,8 @@ bool uActionGlobal::IsBreakingGlobalBind(int sym){
 //    the generic keymaps
 // ***************************
 
-uBind *keymap[SDLK_NEWLAST];
-bool        pressed[SDLK_NEWLAST];
+tJUST_CONTROLLED_PTR< uBind > keymap[SDLK_NEWLAST];
+bool                          pressed[SDLK_NEWLAST];
 
 static void keyboard_init(){
     for(int i=0;i<SDLK_NEWLAST;i++)
@@ -224,7 +226,7 @@ static void keyboard_init(){
 
 static void keyboard_exit(){
     for(int i=0;i<SDLK_NEWLAST;i++)
-        tDESTROY(keymap[i]);
+        keymap[i] = 0;
     //uBindPlayer::Init();
     //global_bind::Init();
 }
@@ -235,12 +237,15 @@ static tInitExit keyboard_ie(&keyboard_init, &keyboard_exit);
 // generic keypress/mouse movement binding class
 // *********************************************
 
-uBind::uBind(std::istream &s): lastValue_(0), delayedValue_(0), act(NULL){
-    char name[30];
+uBind::~uBind(){}
+
+uBind::uBind(uAction *a ):lastValue_(0), delayedValue_(0), lastSym_(-1), lastTime_(-1), act(a){}
+
+uBind::uBind(std::istream &s): lastValue_(0), delayedValue_(0), lastSym_(-1), lastTime_(-1), act(NULL)
+{
+    std::string name;
     s >> name;
-    for(int i=su_allActionsLen-1;i>=0;i--)
-        if(!strcmp(name,su_allActions[i]->internalName))
-            act=su_allActions[i];
+    act = uAction::Find( name.c_str() );
 }
 
 void uBind::Write(std::ostream &s){
@@ -273,8 +278,22 @@ void uBind::HanldeDelayed()
     }
 }
 
+REAL su_doubleBindTimeout=-10.0f;
 
+bool uBind::IsDoubleBind( int sym )
+{
+    double currentTime = tSysTimeFloat();
 
+    // if a different key was used for this action a short while ago, give alarm.
+    bool ret = ( su_doubleBindTimeout > 0 && sym != lastSym_ && currentTime - lastTime_ < su_doubleBindTimeout );
+
+    // store last usage
+    lastSym_ = sym;
+    lastTime_ = currentTime;
+
+    // return result
+    return ret;
+}
 
 // *******************
 // player config
@@ -322,7 +341,7 @@ int uPlayerPrototype::Num(){return nextid;}
 //  Menuitem for input selection
 // *****************************************************
 
-static char *keyname(int sym){
+static char const * keyname(int sym){
 #ifndef DEDICATED
     if (sym<=SDLK_LAST)
         return SDL_GetKeyName(static_cast<SDLKey>(sym));
@@ -402,8 +421,8 @@ public:
         switch (e.type){
         case SDL_MOUSEMOTION:
             if(active){
-                int xrel=e.motion.xrel;
-                int yrel=-e.motion.yrel;
+                REAL xrel=e.motion.xrel;
+                REAL yrel=-e.motion.yrel;
 
                 if (fabs(xrel)>MREL*fabs(yrel)){ // x motion
                     if (xrel>MTHRESH) // left
@@ -438,19 +457,24 @@ public:
                 SDL_keysym &c=e.key.keysym;
                 if(!active){
                     if (c.sym==SDLK_DELETE || c.sym==SDLK_BACKSPACE)
+                    {
                         for(int keysym=SDLK_NEWLAST-1;keysym>=0;keysym--)
                             if(keymap[keysym] &&
                                     keymap[keysym]->act==act &&
                                     keymap[keysym]->CheckPlayer(ePlayer)){
-                                delete keymap[keysym];
                                 keymap[keysym]=NULL;
                             }
+                        return true;
+                    }
                     return false;
                 }
-                if (c.sym!=SDLK_ESCAPE)
-                    sym=c.sym;
 
                 active=0;
+
+                if (c.sym!=SDLK_ESCAPE)
+                    sym=c.sym;
+                else
+                    return true;
             }
             break;
         default:
@@ -461,13 +485,11 @@ public:
             if(keymap[sym] &&
                     keymap[sym]->act==act &&
                     keymap[sym]->CheckPlayer(ePlayer)){
-                delete keymap[sym];
                 keymap[sym]=NULL;
             }
             else{
-                if (keymap[sym])
-                    delete keymap[sym];
-                keymap[sym]=new uBindPlayer(act,ePlayer);
+                keymap[sym]=NULL;
+                keymap[sym]=uBindPlayer::NewBind(act,ePlayer);
             }
             return true;
         }
@@ -511,7 +533,8 @@ static void s_InputConfigGeneric(int ePlayer, uAction *&actions,const tOutput &t
 
     uMenu input_menu(title);
 
-    actions->tListItemBase::Sort(&Input_Compare);
+    if(actions)
+        actions->tListItemBase::Sort(&Input_Compare);
 
     int len = actions->Len();
 
@@ -652,6 +675,8 @@ bool su_HandleEvent(SDL_Event &e, bool delayed ){
         if (keymap[sym]->act->type==uAction::uINPUT_ANALOG)
             pm*=ts*key_sensitivity;
         pressed[sym]=(realpm>0);
+        if ( pm > 0 && keymap[sym]->IsDoubleBind( sym ) )
+            return true;
         return (keymap[sym]->Activate(pm, delayed ));
 
     }
@@ -664,26 +689,69 @@ void su_InputSync(){
     double time=tSysTimeFloat();
     ts=REAL(time-lastTime);
 
-    static REAL tsSmooth=0;
-    tsSmooth+=REAL(ts*.1);
-    tsSmooth/=REAL(1.1);
+    //static REAL tsSmooth=0;
+    //tsSmooth+=REAL(ts*.1);
+    //tsSmooth/=REAL(1.1);
     lastTime=time;
 
     for(int sym=SDLK_NEWLAST-1;sym>=0;sym--)
         if (pressed[sym] && keymap[sym] &&
                 keymap[sym]->act->type==uAction::uINPUT_ANALOG)
-            keymap[sym]->Activate(tsSmooth*key_sensitivity, su_delayed );
+            keymap[sym]->Activate(ts*key_sensitivity, su_delayed );
 }
 
+void su_ClearKeys()
+{
+    for(int sym=SDLK_NEWLAST-1;sym>=0;sym--)
+    {
+        if (pressed[sym] && keymap[sym] )
+            keymap[sym]->Activate(-1, su_delayed );
+        pressed[sym] = false;
+    }
+}
 
 // *****************
 // Player binds
 // *****************
 
-static char *Player_keyword="PLAYER_BIND";
+static char const * Player_keyword="PLAYER_BIND";
 
-uBindPlayer::uBindPlayer(std::istream &s):uBind(s){
-    s >> ePlayer;
+uBindPlayer::uBindPlayer(uAction *a,int p):uBind(a),ePlayer(p){}
+
+uBindPlayer::~uBindPlayer(){}
+
+uBindPlayer * uBindPlayer::NewBind(std::istream &s)
+{
+    // read action
+    std::string actionName;
+    s >> actionName;
+    uAction * act = uAction::Find( actionName.c_str() );
+
+    // read player ID
+    int player;
+    s >> player;
+
+    // delegate
+    return NewBind( act, player );
+}
+
+uBindPlayer * uBindPlayer::NewBind( uAction * action, int player )
+{
+    // see if the bind has an alias
+    for ( int i = SDLK_NEWLAST-1; i >= 0; --i )
+    {
+        // compare action
+        uBind * old = keymap[i];
+        if ( old && old->act == action )
+        {
+            uBindPlayer * oldPlayer = dynamic_cast< uBindPlayer * >( old );
+            if ( oldPlayer && oldPlayer->ePlayer == player )
+                return oldPlayer;
+        }
+    }
+
+    // no alias found, return new bind
+    return tNEW(uBindPlayer)( action, player );
 }
 
 bool uBindPlayer::IsKeyWord(const char *n){
